from flask import Flask, request
from os import path
from urllib.parse import urljoin
import mysql.connector
from mysql.connector import errorcode
import requests
import simplejson
import settings

import xml.etree.ElementTree as ET

# Flask init
app = Flask(__name__, static_url_path='', static_folder='')

# Connect to MySQL
try:
    cnx = mysql.connector.connect(user=settings.DB_USER,
                                  password=settings.DB_PASSWD,
                                  host=settings.DB_HOST,
                                  port=settings.DB_PORT,
                                  database=settings.DB_NAME)

    print("Connected to database: {}".format(settings.DB_NAME))
except mysql.connector.Error as err:
    if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
        print("Something is wrong with your user name or password")
    elif err.errno == errorcode.ER_BAD_DB_ERROR:
        print("Database does not exist")
    else:
        print(err)


def get_type_name(cur_type):
    if cur_type in [0, 1, 2, 3, 4, 5, 6, 8, 9, 246]:
        return 'number'
    elif cur_type in [7, 10, 11, 12]:
        return 'datetime'

    return 'string'


def render_data_table(record_set):

    table = {}
    table['cols'] = list(map(lambda item:
                         {
                            'id': item[0],
                            'label': item[0],
                            'type': get_type_name(item[1])},
                         record_set['fields']))

    def map_row_item(row):
        row_data = {'c': []}

        key_index = 0
        for value in row:
            if table['cols'][key_index]['type'] == 'number':
                row_data['c'].append({'v': value * 1})
            elif table['cols'][key_index]['type'] == 'datetime':
                if value:
                    row_data['c'].append({'v': f'Date({value.year}, {value.month - 1}, {value.day})'})
                else:
                    row_data['c'].append({'v': 'Date(0000,00,00)'})
            else:
                row_data['c'].append({'v': value})

            key_index += 1

        return row_data

    table['rows'] = list(map(lambda item: map_row_item(item), record_set['rows']))

    return table


def render_requested_list(record_set):

    def map_row_item(item):
        id_field = str(item[0])
        if len(item) > 0:
            text_field = item[1]
        else:
            text_field = id_field

        return {'id': id_field, 'text': text_field}

    return list(map(lambda item: map_row_item(item), record_set['rows']))


def execute_sql(sql):
    cursor = cnx.cursor()
    cursor.execute(sql)
    result = cursor.fetchall()
    return {'rows': result, 'fields': cursor.description}


def build_sql(query):
    # send a request to the REST web-service
    request_data = {
        'modelId': settings.MODEL_ID,
        'query': query
    }
    headers = {
                'Content-type': "application/json",
                'SQB-Key': settings.SQBAPI_KEY
    }
    response = requests.post(urljoin(settings.SQBAPI_HOST, 'api/2.0/SqlQueryBuilder'), simplejson.dumps(request_data), headers=headers)
    if response.status_code >= 400:
        print("Error: " + response.text)
        return ""

    result = response.json()

    # now we get an SQL statement by the query defined on client-side
    if result['sql']:
        return result['sql']
    else:
        raise Exception("Error: No 'sql' field")


def get_model_xml(model_id):
    headers = {'SQB-Key': settings.SQBAPI_KEY}
    response = requests.get(urljoin(urljoin(settings.SQBAPI_HOST, 'api/2.0/DataModels/'), model_id), headers=headers)
    return response.text


# REST API part
@app.route('/', methods=['GET'])
def index():
    return app.send_static_file("QueryBuilder.html")


@app.route('/getModel', methods=['POST'])
def get_model():
    # read model from a file and return in response
    model_path = path.join(path.dirname(__file__), settings.MODEL_FILE_JSON)
    with open(model_path, 'r') as file:
        model = file.read()

    return model


@app.route('/loadQuery', methods=['POST'])
def load_query():
    body_json = request.get_json()
    query_id = body_json['id']
    query_path = path.join(path.dirname(__file__), query_id + ".json")

    # read query from a file and return in response
    with open(query_path, 'r') as file:
        query = file.read()

    return query


@app.route('/saveQuery', methods=['POST'])
def save_query():
    body_json = request.get_json()
    query = body_json['query']
    query_name = query['id']

    query_path = path.join(path.dirname(__file__), query_name + ".json")
    with open(query_path, 'w') as file:
        file.write(simplejson.dumps(query))

    return simplejson.dumps({'result': 'OK'})


@app.route('/syncQuery', methods=['POST'])
def sync_query():
    #  return generated SQL to show it on our demo web-page. Not necessary to do in production!
    body_json = request.get_json()
    sql = build_sql(body_json['query'])
    return simplejson.dumps({'statement': sql})


@app.route('/executeQuery', methods=['POST'])
def execute_query():
    body_json = request.get_json()
    sql = build_sql(body_json['query'])
    record_set = execute_sql(sql)
    if record_set:
        result_set = render_data_table(record_set)
        return simplejson.dumps({'statement': sql, 'resultSet': result_set})

    return simplejson.dumps({'statement': "DATABASE CONNECTION ERROR!!!"})


@app.route('/listRequest', methods=['POST'])
def list_request():
    # here  we need to assemble the requested list based on its name and return it as JSON array
    # each item in that array is an object with two properties: "id" and "text"

    # get the name of requested list
    body_json = request.get_json()
    list_name = body_json['listName']

    if list_name == "SQL":
        # if this is a SQL list request - we need to execute SQL statement and return the result
        # set as a list of of {id, text} items

        # First we need to get out xml model from the SQB
        editor_id = body_json['editorId']
        model_xml = get_model_xml(settings.MODEL_ID)

        # We need to parse xml model to find our sql
        model = ET.fromstring(model_xml)
        result = model.findall('.//Editor/[@id="{}"]/SQL'.format(editor_id))

        if result:
            sql = result[0].text
            if sql:
                # And now execute our sql to get data and render list
                record_set = execute_sql(sql);
                if record_set:
                    return simplejson.dumps(render_requested_list(record_set))

    if list_name == "RegionsList":
        return simplejson.dumps(
            [
                {'id': "11", 'text': "AAAA"},
                {'id': "22", 'text': "BBBB"},
                {'id': "33", 'text': "CCCC"},
                {'id': "44", 'text': "DDDD"}
            ]
        )

    return simplejson.dumps([])


if __name__ == '__main__':
    app.run(port=settings.APP_PORT)
