var config = require('./config');

var pg = require('pg');
var fs = require("fs");
var express = require("express");
var bodyParser = require("body-parser");
var request = require("request");
var path = require("path");
var xml2js = require("xml2js");
var urljoin = require('url-join');

var pool = new pg.Pool({
    host     : config.DB_HOST,
    port     : config.DB_PORT,
    user     : config.DB_USER,
    password : config.DB_PASSWD,
    database : config.DB_NAME
});

pool.connect(function(err) {
    if (err) throw err;
    console.log("Connected!");
});

function getTypeName(type) {
    switch (type) {
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 8:
        case 20:
        case 23:
        case 700:
            return 'number';
        case 7:
        case 10:
        case 11:
        case 12:
        case 1082:
            return 'datetime';
        default:
            return 'string';
    }
}

function renderDataTable(recordSet) {
    var table = {};

    table.cols = recordSet.fields.map(item => {
        return {
            id: item.name,
            label: item.name,
            type: getTypeName(item.dataTypeID)
        }
    });

    table.rows = recordSet.rows.map(row => {
        var rowData = {};

        rowData.c = Object.keys(row).map((key, keyIdx) => {
            let value = row[key];

            switch (table.cols[keyIdx].type) {
                case 'number':
                    return { v: value * 1 }
                case 'datetime':
                    if (value !== null){
                        var dateString = value.toISOString();
                        return { v: "Date("+ dateString.substr(0, 4) + ", "+ (parseInt(dateString.substr(5, 2)) - 1) + 
                        ", " + dateString.substr(8, 2) + ")" }
                    }
                    return { v: "Date(0000,00,00)" }
                default:
                    return { v: value }
            }
        });

        return rowData;
    });

    return table;
}

function renderRequestedList(recordSet) {
    let result = recordSet.rows.map(row => {
        let id_field = row[Object.keys(row)[0]];
        let text_field = Object.keys(row).length > 1 ? row[Object.keys(row)[1]] : id_field;
        
        return {
            id: id_field.toString(),
            text: text_field
        };
    });

    return JSON.stringify(result);
}

function executeSql(sql) {

    return new Promise(function(resolve, reject){
        pool.query(sql, (error,result)=>{
            if (error){
                reject(error);
				return;
            }
			if (!result) {
                reject('Empty result');
                return;
            }

			//console.log(result);
            var recordSet = {rows: result.rows,fields: result.fields }
            resolve(recordSet);
        })
    });
}

function buildSql(query) {
    //send a request to the REST web-service
    var request_data = {
        modelId: config.MODEL_ID,
        query: query
    };

    return new Promise(function(resolve, reject){
        let obj = {
            url: urljoin(config.SQBAPI_HOST, 'api/2.0/SqlQueryBuilder'), 
            method: 'POST',
            headers: {
                'Content-type': "application/json",
                'SQB-Key': config.SQBAPI_KEY
             },
            body:  JSON.stringify(request_data)
        }

        request(obj,
		function(error, response, body){
            if (error) {
                reject(error);
            }
            else if (response.statusCode >= 400) {
                reject(new Error(response.body));
            }
            resolve(response.body);
        })  
    })
    .then((resJson) => {
        //get a response in JSON format	
        try {         
            var resObj = JSON.parse(resJson); 	
    
            var sql = "";
            //now we get an SQL statement by the query defined on client-side
            if (resObj.sql !== null) {
                sql = resObj.sql;
            }
            else {
                throw new Error("Error: No 'sql' field");
            }
    
            return sql;
        }
        catch (e) {
            throw new Error("Error: Invalid response: " + e);
        }
    });
}

function getXmlModel(modelId){
    return new Promise(function (resolve, reject){
        request({
            url: urljoin(config.SQBAPI_HOST, 'api/2.0/DataModels', modelId), 
            method: 'GET',
            headers: {
                'SQB-Key': config.SQBAPI_KEY
             }
        },function(error, response, body){ 
          if (error){
              reject(error)
          } else if (response.statusCode >= 400){
              reject(new Error("Can't load xml"))
          }
          resolve(response.body);
        })  
    });
}




var app = express();

app.use(bodyParser.json());

app.get('/', function(request, response){
    response.sendFile(path.join(__dirname, 'QueryBuilder.html')); 
});

app.post('/getModel', function(request, response){
    var model_name = request.body.modelId;
    
    //read model from a file and return in response

    fs.readFile(path.join(__dirname, config.MODEL_FILE_JSON),'utf8', function(error, model){
        response.send(model);
      //  console.log(model);
    });
   
});

app.post('/loadQuery', function(request, response){
    //get query id
    var query_id = request.body.id;
     
    //read query from a file and return in response
    var query_file_name = query_id + ".json";
    fs.readFile(path.join(__dirname, query_file_name), 'utf8', function(error, query_json){
        response.send(query_json);
     //   console.log(query_json);
    });
     
});

app.post('/saveQuery', function(request, response){
    //get query 
    var query = request.body.query;
     
      //get query name(id)
    var query_name = query.id;
     
    //save query to a file in json format
    var query_file_name = query_name + ".json";
    fs.writeFile(path.join(__dirname, query_file_name), JSON.stringify(query), 'utf8');

    response.send('{"result":"OK"}');
         
  //  console.log('{"result":"OK"}');
});


app.post('/syncQuery', function(request, response){
    //return generated SQL to show it on our demo web-page. Not necessary to do in production!
    buildSql(request.body.query).then((sql)=>{
        var result = JSON.stringify({statement: sql});
        response.send(result);
        //    console.log(result);
    })   
    .catch((error) => {
        console.log(error);
        response.statusCode = 400;
        response.send("Error: " + error);
    });	
});

app.post('/executeQuery', function(request, response){
    buildSql(request.body.query).then((sql) => {
        var result = '{}';
        executeSql(sql).then((recordSet) => {
            if (recordSet){
                var resultSet = renderDataTable(recordSet);
                result = JSON.stringify({statement: sql, resultSet: resultSet });	
            }
            else {
                result = JSON.stringify({statement: "DATABASE CONNECTION ERROR!!!"});		
            }
            
            response.send(result);
         //   console.log(result);
        }).catch((error) => {
            response.statusCode = 400;
            response.send(error);
        });
    }).catch((error) => {
        response.statusCode = 400;
        response.send(error);
    });
});

app.post('/listRequest', function(request, response){
    //here  we need to assemble the requested list based on its name and return it as JSON array
    //each item in that array is an object with two properties: "id" and "text"

    //get the name of requested list
    var list_name = request.body.listName;

    if (list_name === "SQL") {
        //if this is a SQL list request - we need to execute SQL statement and return the result set as a list of of {id, text} items
        //First we need to get out xml model from the SQB
        getXmlModel(config.MODEL_ID).then((xmlModel) =>{

            //We need to parse xml model to find our sql
            //Here is an example using xml2js framework
            xml2js.parseString(xmlModel, function(error, result){
                try {
                    if (error) throw error;
                        
                    let editorList = result.DataModel.Editors[0].Editor;

                    let editor = editorList.find(item => {
                        return item['$'].id === request.body.editorId;
                    });

                    let sql = "";
                    if (editor) {
                        sql = editor.SQL[0];
                    }
                    else {
                        response.send('[]');
                        return;
                    }
            
                    //And now execute our sql to get data and render list
                    executeSql(sql).then((recordSet) => {
                        if(recordSet !== null) {
                            var result = renderRequestedList(recordSet);
                            response.send(result);
                        }
                    });
                } catch (e){
                    response.statusCode(400).send("Error: " + e);
                }
            });  
           
        }).catch((error) => {
            response.statusCode(400).send("Error: " + error);
        });
    }
    else {
        //otherwise we return some list based on list name
        if (list_name === "RegionsList") {
            response.send(JSON.stringify([
                {id:"11", text:"AAAA"},
                {id:"22", text:"BBBB"},
                {id:"33", text:"CCCC"},
                {id:"44", text:"DDDD"}
            ]));
        }
        else {
            response.send('[]');
        }		
    }
});

var port = process.env.PORT || config.port || 3200; 
app.listen(port);

console.log("Server run on localhost: ", port);